import { createPopper } from "@popperjs/core";


/*
 * Choice JS
*/

class Builder {



    //Constructor
    constructor(target, userOptions){
        this.target = target;
        this.dropdown = false;
        //Build our default options
        this.options = {
            ...this.buildDefaultOptions(userOptions),
            ...userOptions
        };
        this.closeListener = this.closeListener.bind(this);
    }



    /**
     * Create a new choice dropdown
     *
     * @return {Promise}
    */
    async create(){
        //Build the select box
        this.choice = await this.buildChoice();
        //Global bindings
        await this.bindActiveListener();
        this.bindEventListener();
        //Dispatch event to load it
        this.target.dispatchEvent(new Event('change'));
        //Return
        return this.choice;
    }



    /**
     * Build the default options
     *
     * @param {Object} userOptions The user options for any overwrites
     * @return {Object} An object of options
    */
    buildDefaultOptions(userOptions){
        let options = {
            width:'inherit',
            height:'inherit',
            placeholder: (this.target.hasAttribute('placeholder') ? this.target.getAttribute('placeholder') : false),
            class: '',
            closeOnSelect: false,
            event: 'change',
            copyAttributes: true,
            bindTo: 'parent'
        }
        //Return
        return options;
    }



    /**
     * Configure options for the choice
     *
     * @return {Promise}
    */
    async configureOptions(choice){

        //Width and height
        choice.style.width = (this.options.width == 'inherit') ? (this.target.offsetWidth + 'px') : this.options.width,
        choice.style.minHeight = (this.options.height == 'inherit') ? (this.target.offsetHeight + 'px') : this.options.height;

        //Class
        if(this.options.class != '') choice.classList.add(this.options.class);

        //Placeholder
        if(this.options.placeholder) choice.querySelector('.choicejs-select').innerHTML = this.options.placeholder;

        //Return
        return;
    }



    /**
     * Bind active click
     *
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindActiveListener(){

        //Listeners
        this.choice.addEventListener('click', async () => {
            //If we're already open, do nothing
            if(this.choice.classList.contains('active')) return;
            //Create the dropdown
            await this.buildDropdown();
            //Observe for visibility
            this.visibilityObserver.observe(this.choice);
            //Bind close listener
            document.addEventListener('click', this.closeListener);
            //Add the active class
            this.choice.classList.add('active');
            //If the user has provided a search url and the `data-choice-search-preload` attribute, run a keyup to pre-load some data
            if(this.target.hasAttribute('data-choice-search-url') && this.target.hasAttribute('data-choice-search-preload')){
                this.dropdown.querySelector('.choicejs-search').dispatchEvent(new Event('keyup'));
                this.dropdown.querySelector('ul').classList.add('loading');
            }
        })

        //Return
        return;
    }



    /**
     * Bind Event Listener
     *
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindEventListener(){
        //When an event is fired on the target, change our dropdowns
        this.target.addEventListener(this.options.event, async () => {

            //Rebuild the options and pills on the target change
            let multiple = (this.target.hasAttribute('multiple') && ['true', 'multiple'].includes(this.target.getAttribute('multiple'))),
                choice_select = this.choice.querySelector('.choicejs-select');

            //Reset the select placeholder
            choice_select.innerHTML = (this.options.placeholder ? this.options.placeholder : '');

            //If the dropdown is currently open reset all options
            if(this.dropdown){
                for(let option of this.dropdown.querySelectorAll(`ul li`)){
                    option.removeAttribute('selected');
                }
            }

            //Are we working with multiple?
            if(!multiple){

                //Get the value of the selected select option and make sure there is one
                let selected_value = this.target.value;
                if(selected_value != ''){
                    //Set the visual HTML of the choice select
                    choice_select.innerHTML = this.target.querySelector('option:checked').innerHTML;
                    //If theres a dropdown open, enable the option
                    if(this.dropdown){
                        this.dropdown.querySelector(`ul li[value="${selected_value}"]`).setAttribute('selected', true);
                    }
                }

            } else {

                //Start building pills
                let pills = [];
                //Get all of the selected options and check we have some
                let selected_options = this.target.querySelectorAll('option:checked');
                if(selected_options.length > 0){
                    //Reset the choice select innerHTML
                    choice_select.innerHTML = '';
                    //Loop over all of the selected options
                    for(let option of selected_options){
                        //Add a pillbox
                        pills.push(this.buildOptionPill(option));
                        //If there is a dropdown open, enable the option
                        if(this.dropdown){
                            this.dropdown.querySelector(`ul li[value="${option.value}"]`).setAttribute('selected', true);
                        }
                    }
                }
                await Promise.all(pills);

            }
        });
    }



    /**
     * Bind option select
     *
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindOptionSelect(){
        //Loop and add
        for(let option of this.dropdown.querySelectorAll('ul li')){
            //Add the event listener
            option.addEventListener('click', async(e) => {
                //Stop Prop
                e.stopPropagation();
                e.stopImmediatePropagation();
                //Add or remove
                if(option.hasAttribute('selected')){
                    //Remove the option
                    await this.removeOption(option);
                } else {
                    //Add the option
                    await this.addOption(option);
                }
                //Close on select?
                if(this.options.closeOnSelect) this.destroyDropdown();
                //What parent event to trigger?
                if(this.options.event !== false){
                    this.target.dispatchEvent(new Event(this.options.event));
                }
            });
        }
        //Return
        return;
    }



    /**
     * Bind dynamic search if the user has provided an endpoint to use
     *
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindDynamicSearch(){
        //Has the user also asked for multiple? If yes, warn and quit
        let multiple = (this.target.hasAttribute('multiple') && ['true', 'multiple'].includes(this.target.getAttribute('multiple')));
        if(multiple){
            console.info("Choice JS - Multiple options is not supported with live searching.");
            return;
        }
        //Get the url endpoint if we have one
        if(this.target.hasAttribute('data-choice-search-url')){
            //Get the endpoint we're using
            let endpoint = this.target.getAttribute('data-choice-search-url'),
                searchBar = this.dropdown.querySelector('.choicejs-search'),
                choiceList = this.dropdown.querySelector('ul');
            //Watch for the search change
            let searchTimeout = 0;
            //Remove the bindings to avoid duplicates
            searchBar.addEventListener('keyup', _listener => {
                //Try
                try {
                    clearTimeout(searchTimeout);
                    //Set timeout
                    searchTimeout = setTimeout(async () => {
                        //Clear the select options and the choice options, replace this a loader
                        choiceList.innerHTML = '';
                        this.target.innerHTML = '';
                        //Loading
                        choiceList.classList.add('loading');
                        //Get the user input, allow empty so the user can check in backend if empty and return default set
                        let userInput = searchBar.value;
                        //Build our params
                        let url = new URL(endpoint)
                        url.search = new URLSearchParams({choicesearch: userInput}).toString();
                        //Run the search
                        let response = await fetch(url);
                        let result = await response.json();
                        //Remove the loader
                        choiceList.classList.remove('loading');
                        //Add options
                        for(let option of result){
                            //Create a new item
                            let choiceItem = document.createElement('li'),
                                selectItem = document.createElement('option');
                            //Set the visible value
                            choiceItem.innerHTML = option.label;
                            selectItem.innerHTML = option.label;
                            //Set the actual value
                            choiceItem.setAttribute('value', option.value);
                            selectItem.value = option.value;
                            //Append it to our select
                            this.target.appendChild(selectItem);
                        }


                        //Build the options
                        await this.buildOptionList(choiceList);

                        //Bindings
                        await Promise.all([
                            this.bindOptionSelect()
                        ]);


                    }, 1000);
                } catch(e){
                    console.log('ChoiceJS error fetching endpoint', e);
                }
            })
        }
        //Return
        return;
    }



    /**
     * Bind option filtering. When a user types into the search box
     *
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindOptionFiltering(){
        //If we have a dynamic url, do not bind to the list
        if(this.target.hasAttribute('data-choice-search-url')) return;
        //When the choice search bar changes
        var searchTimeout = 0,
            searchBar = this.dropdown.querySelector('.choicejs-search'),
            options = this.dropdown.querySelectorAll('ul li');
        //When the postcode is changed, test it
        searchBar.addEventListener('input', () => {
            //Clear the above timeout
            clearTimeout(searchTimeout);
            //Set a new timeout
            searchTimeout = setTimeout(() =>{
                //Loop all options
                for(let option of options){
                    if(option.innerHTML.toLowerCase().includes(searchBar.value.toLowerCase())){
                        option.classList.remove('hide');
                    } else {
                        option.classList.add('hide');
                    }
                }
            }, 500);
        });
        //Return
        return;
    }



    /* ------------------------------------------------------------------------------------------------------------------------------------------------
    |  HTML Builders
    ------------------------------------------------------------------------------------------------------------------------------------------------ */



    /**
     * Build the choice
     *
     * @return {Promise} HTMLElement The new choice
    */
    async buildChoice(){

        //Create a new id
        this.id = 'choice-'+Math.random().toString(36).substr(2, 4);

        //Build the new Select HTML element and add it directly after the target
        let wrapper = document.createElement('div'),
            select = document.createElement('div');

        //Set attributes
        wrapper.classList.add('choicejs');
        wrapper.setAttribute('id', this.id);
        select.classList.add('choicejs-select');

        //Set the id on the target so we can relate them together
        this.target.setAttribute('data-choicejs-id', `#${this.id}`);

        //Copy the select boxes attributes
        if(this.options.copyAttributes === true){
            //Only copy over limited attributes
            wrapper.setAttribute('name', this.target.getAttribute('name'));
        }

        //Add the new select element to the wrapper
        wrapper.appendChild(select);

        //Configure the options for it
        await this.configureOptions(wrapper);

        //Deploy the select
        this.target.parentNode.insertBefore(wrapper, this.target.nextSibling);

        //Return
        return wrapper;

    }



    /**
     * Build dropdown
     *
     * @returns {Promise}
    */
    async buildDropdown(){

        //Build the new dropdown html elements
        let dropdownWrapper = document.createElement('div'),
            dropdown = document.createElement('div'),
            list = document.createElement('ul'),
            searchBox = document.createElement('input');

        //Set attributes
        searchBox.classList.add('choicejs-search');
        searchBox.setAttribute('placeholder', 'Search');
        dropdownWrapper.classList.add('choicejs-dropdown-wrapper');
        dropdownWrapper.setAttribute('data-attached-choicejs-id', this.id),
        dropdown.classList.add('choicejs-dropdown');

        //Build the whole select and return it
        dropdown.appendChild(searchBox);
        dropdown.appendChild(list);
        dropdownWrapper.appendChild(dropdown);

        //Add the dropdown to the class for easy access
        this.dropdown = dropdownWrapper;

        //Figure out its positioning and style
        dropdown.style.width = `${this.choice.offsetWidth}px`;

        //Build the options
        await this.buildOptionList(list);

        //Where are we attaching the fixed popup?
        switch(this.options.bindTo){
            case 'parent':
                this.choice.appendChild(dropdownWrapper);
            break;
            case 'body':
                document.querySelector('body').appendChild(dropdownWrapper);
            break;
        }

        //Tether
        createPopper(this.choice, dropdownWrapper, {
            placement: 'bottom',
            strategy: 'fixed',
            modifiers: [
                {
                    name: 'offset',
                    options: {
                        offset: [0, 10],
                    },
                },
            ]
        });


        //Bindings
        await Promise.all([
            this.bindOptionSelect(),
            this.bindOptionFiltering(),
            this.bindDynamicSearch()
        ]);

        //Return
        return dropdownWrapper;
    }


    /**
     * Destroy dropdown
     *
     * @returns {Promise}
    */
    destroyDropdown(){
        document.removeEventListener('click', this.closeListener);
        this.dropdown.remove();
        this.visibilityObserver.unobserve(this.choice);
        this.choice.classList.remove('active');
        this.dropdown = false;
    }



    /**
     * Build dropdown
     *
     * @returns {Promise}
    */
    async buildOptionList(dropdownList){

        //Add option function
        let addOption = async (option, wrapper) => {
            let item = document.createElement('li');
            item.innerHTML = option.innerHTML;
            //Copy all attributes over from the option
            if(this.options.copyAttributes === true){
                for (let i = 0; i < option.attributes.length; i++){
                    item.setAttribute(option.attributes[i].nodeName, option.attributes[i].nodeValue);
                }
            }
            //Active?
            if(option.selected == true) item.setAttribute('selected', true);
            wrapper.appendChild(item);
        }

        //Are we working with optgroups?
        let optgroups = this.target.querySelectorAll('optgroup');
        if(optgroups.length > 0){
            for(let optgroup of optgroups){
                let group = document.createElement('div');
                group.classList.add('optgroup');
                group.setAttribute('label', optgroup.getAttribute('label'));
                //Add the options to the opt group
                for(let option of optgroup.querySelectorAll('option')){
                    if(option.value != '' && option.value != undefined) addOption(option, group);
                }
                //Add the group to the list
                dropdownList.appendChild(group);
            }
        } else {
            //Loop the targets options and build our list
            for(let option of this.target.querySelectorAll('option')){
                if(option.value != '' && option.value != undefined) addOption(option, dropdownList);
            }
        }

    }


    /**
     * Build option pill
     *
     * @param {Node} option - The option selected
     * @param {Node} select - The select we're appending to
     * @returns {Promise}
    */
    async buildOptionPill(option){
        //Create the pillbox
        let wrapper = document.createElement('div'),
            close = document.createElement('span'),
            value = option.value;
        //Attributes
        wrapper.classList.add('choicejs-pill');
        wrapper.setAttribute('value', value);
        wrapper.innerHTML = option.innerHTML;
        //Build it
        wrapper.appendChild(close);
        this.choice.querySelector('.choicejs-select').appendChild(wrapper);
        //Add the remove event listener
        close.addEventListener('click', async (e) => {
            //Run remove btn
            await this.removeOption(option);
            //What parent event to trigger?
            if(this.options.event !== false){
                this.target.dispatchEvent(new Event(this.options.event));
            }
        })
        //Return
        return;
    }


    /* ------------------------------------------------------------------------------------------------------------------------------------------------
    |  Misc Functions
    ------------------------------------------------------------------------------------------------------------------------------------------------ */



    /**
     * Add option
     *
     * @param {HTMLElement} option - The option we're selecting
     * @returns {Promise}
    */
    async addOption(option){
        //Get elems
        let value = option.getAttribute('value'),
            multiple = (this.target.hasAttribute('multiple') && ['true', 'multiple'].includes(this.target.getAttribute('multiple')));
        //Mark the option as selected
        this.target.querySelector(`option[value="${value}"]`).selected = true;
        //If its not multiple, set the value of the parent select box
        if(!multiple) this.target.value = value;
    }



    /**
     * Remove option
     *
     * @param {HTMLElement} option - The option we're selecting
     * @returns {Promise}
    */
    async removeOption(option){
        //Get elems
        let value = option.value,
            multiple = (this.target.hasAttribute('multiple') && ['true', 'multiple'].includes(this.target.getAttribute('multiple')));
        //Remove the hidden option
        this.target.querySelector(`option[value="${value}"]`).selected = false;
        this.target.querySelector(`option[value="${value}"]`).removeAttribute('selected');
        //Single or multiple select?
        if(multiple){
            //If there is still options selected, return now, don't reset the value and index of the select
            if(this.target.querySelectorAll(`option:checked`).length > 0) return;
        }
        //Reset
        this.target.value = '';
        this.target.selectedIndex = -1;
    }



    /**
     * Close element on click function
     *
    */
    closeListener(event) {
        if (!this.choice.contains(event.target) && !this.dropdown.contains(event.target)) {
            //Destroy the dropdown
            this.destroyDropdown();
        }
    };



    /**
     * Block observer
    */
    visibilityObserver = new IntersectionObserver((entries, self) => {
        //For each of the entries
        entries.forEach(entry => {
            //If the image is intersecting
            if(!entry.isIntersecting){
                //Hide
                this.destroyDropdown();
            }
        })
    }, {});

}


/**
 * Export the choice function to use
 *
 * @param {NodeElement} target - The target we're adding the choice to
 * @param {Object} options - The options
 * @returns {Promise}  - An instance of choice
 */
export const Choice = async (target, options) => {
    //Start a new Choice
    let newBuild = new Builder(target, options);
    //Run the build
    return await newBuild.create();
}


/**
 * Remove the choice by its target
 *
 * @param {NodeElement} target - The target we're removing the choice from
 */
export const DestroyChoice = async (target) => {
    //If the target doesnt have the data-choicejs-id attribute, return
    if(!target.hasAttribute('data-choicejs-id')) return false;
    //Remove the choice
    this.destroyDropdown();
    target.parentNode.querySelector('.choicejs').remove();
    //Clear and replace the select
    let new_element = target.cloneNode(true);
    //Remove extra stuff added
    new_element.removeAttribute('data-choicejs-id');
    //Replace it
    target.parentNode.replaceChild(new_element, target);
    //Resolve
    return true;
}